#! /usr/bin/env python3

from vasppy.xdatcar import Xdatcar
import argparse
import copy

def parse_command_line_arguments():
    # command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument( 'xdatcar' )
    args = parser.parse_args()
    return( args )

def main():
    args = parse_command_line_arguments()
    xdatcar = Xdatcar()
    xdatcar.read_from( args.xdatcar )
    poscar1 = copy.deepcopy( xdatcar.poscar[0] )
    for p in xdatcar.poscar[1:]:
        poscar2 = p
        for i, c in enumerate( poscar1.coordinates ):
            print( ' '.join( [ str(e) for e in ( poscar1.cell.minimum_image( poscar1.coordinates[i], poscar2.coordinates[i] ).dot( poscar1.cell.matrix ) ) ] ) )
        poscar1 = copy.deepcopy( p )

if __name__ == "__main__":
    main()



import numpy as np  # type: ignore
from scipy.ndimage.filters import gaussian_filter1d  # type: ignore
from pymatgen import Structure  # type: ignore
from typing import List, Optional, TypeVar, Type

"""
This module provides classes for calculating radial disitrbution functions
and Van Hove correlation functions.
"""
RDF = TypeVar('RDF', bound='RadialDistributionFunction')


class RadialDistributionFunction(object):
    """
    Class for computing radial distribution functions.

    Attributes:
        nbins (int): Number of bins.
        range ((float, float)): Minimum and maximum values of r.
        intervals (np.array(float)): r values of the bin edges.
        dr (float): bin width.
        r (float): mid-points of each bin.
        rdf (np.array(float)): RDF values.
        coordination_number (np.array(float)): Volume integral of the RDF.

    """

    def __init__(self,
                 structures: List[Structure],
                 indices_i: List[int],
                 indices_j: Optional[List[int]] = None,
                 nbins: int = 500,
                 r_min: float = 0.0,
                 r_max: float = 10.0,
                 weights: Optional[List[float]] = None) -> None:
        """
        Initialise a RadialDistributionFunction instance.

        Args:
            structures (list(pymatgen.Structure)): List of pymatgen Structure objects.
            indices_i (list(int)): List of indices for species i.
            indices_j (:obj:`list(int)`, optional): List of indices for species j. Optional,
                default is `None`.
            nbins (:obj:`int`, optional): Number of bins used for the RDF. Optional, default is 500.
            rmin (:obj:`float`, optional): Minimum r value. Optional, default is 0.0.
            rmax (:obj:`float`, optional): Maximum r value. Optional, default is 10.0.
            weights (:obj:`list(float)`, optional): List of weights for each structure.
                Optional, default is `None`.

        Returns:
             None

        """
        if weights:
            if len(weights) != len(structures):
                raise ValueError('List of structure weights needs to be the same length'
                                 ' as the list of structures.')
        else:
            weights = [1.0] * len(structures)
        self.self_reference = (not indices_j) or (indices_j == indices_i)
        if not indices_j:
            indices_j = indices_i
        self.indices_i = indices_i
        self.indices_j = indices_j
        self.nbins = nbins
        self.range = (r_min, r_max)
        self.intervals = np.linspace(r_min, r_max, nbins + 1)
        self.dr = (r_max - r_min) / nbins
        self.r = self.intervals[:-1] + self.dr / 2.0
        ff = shell_volumes(self.intervals)
        self.coordination_number = np.zeros(nbins)
        self.rdf = np.zeros((nbins), dtype=np.double)
        for structure, weight in zip(structures, weights):
            hist = np.histogram(self.__dr_ij(structure),
                                bins=nbins,
                                range=(r_min, r_max),
                                density=False)[0]
            rho = float(len(self.indices_i)) / structure.lattice.volume
            self.rdf += hist * weight / rho
            self.coordination_number += np.cumsum(hist)
        self.rdf = self.rdf / ff / sum(weights) / float(len(indices_j))
        self.coordination_number = self.coordination_number / \
            sum(weights) / float(len(self.indices_j))

    def smeared_rdf(self, 
                    sigma: float = 0.1) -> np.ndarray:
        """
        Smear the RDF with a Gaussian kernel.

        Args:
            sigma (:obj:`float`, optional): Standard deviation for Gaussian kernel. 
                Optional, default is 0.1.

        Returns:
            (np.array): Smeared RDF data.

        """
        sigma_n_bins = sigma / self.dr
        return gaussian_filter1d(self.rdf, sigma=sigma_n_bins)

    @classmethod
    def from_species_strings(cls: Type[RDF], 
                             structures: List[Structure], 
                             species_i: str, 
                             species_j: Optional[str] = None, 
                             **kwargs) -> RDF:
        """
        Initialise a RadialDistributionFunction instance by specifying species strings.

        Args:
            structures (list(pymatgen.Structure)): List of pymatgen Structure objects.
            species_i (str): String for species i, e.g. ``"Na"``.
            species_j (:obj:`str`, optional): String for species j, e.g. ``"Cl"``. Optional,
                default is `None`. 
            **kwargs: Variable length keyword argument list. 
                See :func:`vasppy.rdf.RadialDistributionFunction`
                for the full list of accepted arguments.

        Returns:
            (RadialDistributionFunction)

        """
        indices_i: List[int]
        indices_j: Optional[List[int]]

        indices_i = [i for i, site in 
                     enumerate(structures[0]) if site.species_string is species_i]
        if species_j:
            indices_j = [j for j, site in 
                         enumerate(structures[0]) if site.species_string is species_j]
        else:
            indices_j = None
        return cls(structures=structures, 
                   indices_i=indices_i, 
                   indices_j=indices_j, 
                   **kwargs)

    def __dr_ij(self, 
                structure: Structure) -> np.ndarray:
        """
        Calculate all i-j interatomic distances for a single pymatgen Structure.

        Args:
            structure (:obj:`pymatgen.Structure`): A pymatgen Structure.

        Returns:
            np.array: 1D numpy array of length N_i x N_j of distances.

        """
        lattice = structure.lattice
        i_frac_coords = structure.frac_coords[self.indices_i]
        j_frac_coords = structure.frac_coords[self.indices_j]
        dr_ij = lattice.get_all_distances(i_frac_coords, j_frac_coords)
        # Mask dr_ij 2D array to remove i==j dr=0 terms
        mask = np.ones(dr_ij.shape, dtype=bool)
        if self.self_reference:
            np.fill_diagonal(mask, 0)
        return np.ndarray.flatten(dr_ij[mask])


VHA = TypeVar('VHA', bound='VanHoveAnalysis')


class VanHoveAnalysis(object):
    """
    Class for computing Van Hove correlation functions.

    Attributes:
        nbins (int): Number of bins.
        range ((float, float)): Minimum and maximum values of r.
        intervals (np.array(float)): r values of the bin edges.
        dr (float): bin width.
        r (float): mid-points of each bin.
        gsrt (np.array(float)): Self part of the Van Hove correlation function.
        gdrt (np.array(float)): Distinct part of the Van Hove correlation function.

    """

    def __init__(self,
                 structures: List[Structure],
                 indices: List[int],
                 d_steps: int,
                 nbins: int = 500,
                 r_min: float = 0.0,
                 r_max: float = 10.0):
        """
        Initialise a VanHoveCorrelationFunction instance.

        Args:
            structures (list(pymatgen.Structure)): List of pymatgen Structure objects.
            indices (list(int)): List of indices for species to consider.
            d_steps (int): number of steps between structures at dt=0 and dt=t.
            nbins (:obj:`int`, optional): Number of bins used for the RDF. Optional, default is 500.
            rmin (:obj:`float`, optional): Minimum r value. Optional, default is 0.0.
            rmax (:obj:`float`, optional): Maximum r value. Optional, default is 10.0.

        Returns:
             None

        """
        self.nbins = nbins
        self.range = (r_min, r_max)
        self.intervals = np.linspace(r_min, r_max, nbins + 1)
        self.dr = (r_max - r_min) / nbins
        self.r = self.intervals[:-1] + self.dr / 2.0
        self.gdrt = np.zeros((nbins), dtype=np.double)
        self.gsrt = np.zeros((nbins), dtype=np.double)
        rho = len(indices) / structures[0].lattice.volume
        lattice = structures[0].lattice
        ff = shell_volumes(self.intervals)
        rho = len(indices) / lattice.volume
        for struc_i, struc_j in zip(structures[:len(structures) - d_steps], structures[d_steps:]):
            i_frac_coords = struc_i.frac_coords[indices]
            j_frac_coords = struc_j.frac_coords[indices]
            dr_ij = lattice.get_all_distances(i_frac_coords, j_frac_coords)
            mask = np.ones(dr_ij.shape, dtype=bool)
            np.fill_diagonal(mask, 0)
            distinct_dr_ij = np.ndarray.flatten(dr_ij[mask])
            hist = np.histogram(distinct_dr_ij, bins=nbins,
                                range=(0.0, r_max), density=False)[0]
            self.gdrt += hist / rho
            self_dr_ij = np.ndarray.flatten(dr_ij[np.invert(mask)])
            hist = np.histogram(self_dr_ij, bins=nbins,
                                range=(0.0, r_max), density=False)[0]
            self.gsrt += hist / rho
        self.gdrt = self.gdrt / ff / \
            (len(structures) - d_steps) / float(len(indices))
        self.gsrt = self.gsrt / \
            (len(structures) - d_steps) / float(len(indices))

    def self(self, 
             sigma: Optional[float] = None) -> np.ndarray:
        if sigma:
            return self.smeared_gsrt(sigma=sigma)
        else:
            return self.gsrt

    def distinct(self,
                 sigma: Optional[float] = None) -> np.ndarray:
        if sigma:
            return self.smeared_gdrt(sigma=sigma)
        else:
            return self.gdrt

    def smeared_gsrt(self, 
                     sigma: float = 0.1) -> np.ndarray:
        """
        Smear the self part of the Van Hove correlation function with a Gaussian kernel.

        Args:
            sigma (:obj:`float`, optional): Standard deviation for Gaussian kernel. Optional, default is 0.1.

        Returns:
            (np.array): Smeared data.

        """
        sigma_n_bins = sigma / self.dr
        return gaussian_filter1d(self.gsrt, sigma=sigma_n_bins)

    def smeared_gdrt(self,
                     sigma: float = 0.1) -> np.ndarray:
        """
        Smear the distinct part of the Van Hove correlation function with a Gaussian kernel.

        Args:
            sigma (:obj:`float`, optional): Standard deviation for Gaussian kernel. Optional, default is 0.1.

        Returns:
            (np.array): Smeared data.

        """
        sigma_n_bins = sigma / self.dr
        return gaussian_filter1d(self.gdrt, sigma=sigma_n_bins)


def shell_volumes(intervals: np.ndarray) -> np.ndarray:
    """Volumes of concentric spherical shells.

    Args:
        intervals (np.array): N radial boundaries used to define the set of N-1 shells.

    Returns:
        np.array: Volumes of each shell.

    """
    return 4.0 / 3.0 * np.pi * (intervals[1:]**3 - intervals[:-1]**3)



# Summary class and helper methods
# Used for summarising VASP calculations as YAML

from pymatgen.io.vasp.outputs import Vasprun  # type: ignore
from pymatgen.analysis.transition_state import NEBAnalysis  # type: ignore
from vasppy.vaspmeta import VASPMeta
from vasppy.outcar import final_energy_from_outcar, vasp_version_from_outcar, potcar_eatom_list_from_outcar
from vasppy.data.potcar_data import potcar_md5sum_data, potcar_nelect
from vasppy.utils import file_md5, md5sum, match_filename, cd
from xml.etree import ElementTree as ET 
import sys
import yaml
import glob
import re

potcar_sets = ['PBE', 'PBE_52', 'PBE_54', 'PBE_54r', 'LDA_54r', 
               'LDA', 'LDA_52', 'LDA_54', 'GGA', 'USPP_GGA', 'USPP_LDA']

def load_vasp_summary( filename ):
    """
    Reads a `vasp_summary.yaml` format YAML file and returns
    a dictionary of dictionaries. Each YAML document in the file
    corresponds to one sub-dictionary, with the corresponding
    top-level key given by the `title` value.

    Example:
        The file::
    
            ---
            title: foo
            data: foo_data
            ---
            title: bar
            data: bar_data

        is converted to the dictionary::

            { 'foo': { 'title': 'foo', 'data': 'foo_data' },
              'bar': { 'title': 'bar', 'data': 'bar_data' } }

    Args:
        filename (str): File path for the `vasp_summary.yaml` file.

    Returns:
        (dict(dict,dict,...)): A dictionary of separate YAML documents,
            each as dictionaries.a

    """
    with open( filename, 'r' ) as stream:
        docs = yaml.load_all( stream, Loader=yaml.SafeLoader )
        data = { d['title']: d for d in docs }
    return data

def potcar_spec(filename, return_hashes=False):
    """
    Returns a dictionary specifying the pseudopotentials contained in a POTCAR file.

    Args:
        filename (str): The name of the POTCAR file to process.
        return_hash (bool): If True the return dictionary values will be the md5 hashes of
            the component pseudopotential files.

    Returns:
        (Dict): A dictionary of pseudopotential filename: dataset pairs, e.g.
                {'Fe_pv': 'PBE_54', 'O', 'PBE_54'}
    """
    p_spec = {}
    with open(filename, 'r') as f:
        potcars = re.split('(End of Dataset\n)', f.read() )
    potcar_md5sums = [md5sum(''.join(pair)) for pair in zip(potcars[::2], potcars[1:-1:2])]
    for this_md5sum in potcar_md5sums:
        for ps in potcar_sets:
            for p, p_md5sum in potcar_md5sum_data[ps].items():
                if this_md5sum == p_md5sum:
                    if return_hashes:
                        p_spec[p] = this_md5sum
                    else:
                        p_spec[p] = ps
    if len(p_spec) != len(potcar_md5sums):
        raise ValueError( 'One or more POTCARs did not have matching md5 hashes' )
    return p_spec
  
def find_vasp_calculations():
    """
    Returns a list of all subdirectories that contain either a vasprun.xml file
    or a compressed vasprun.xml.gz file.

    Args:
        None

    Returns:
        (List): list of all VASP calculation subdirectories.
    """
    dir_list = [ './' + re.sub( r'vasprun\.xml', '', path ) for path in glob.iglob( '**/vasprun.xml', recursive=True ) ]
    gz_dir_list = [ './' + re.sub( r'vasprun\.xml\.gz', '', path ) for path in glob.iglob( '**/vasprun.xml.gz', recursive=True ) ]
    return dir_list + gz_dir_list

class Summary:
    """
    TODO Document Summary class
    """

    supported_flags = { 'title': 'Title',
                        'description': 'Description',
                        'notes': 'Notes',
                        'type': 'Type',
                        'status': 'Status',
                        'stoichiometry': 'Stoichiometry',
                        'potcar': 'POTCAR',
                        'eatom': 'POTCAR EATOM values',
                        'plus_u': 'Dudarev +U parameters',
                        'energy': 'Energy',
                        'lreal': 'LREAL',
                        'k-points': 'k-points',
                        'functional': 'functional',
                        'encut': 'encut',
                        'ediffg': 'ediffg',
                        'ibrion': 'ibrion',
                        'converged': 'converged',
                        'md5': 'md5',
                        'directory': 'directory',
                        'vbm': 'Vasprun valence band maximum',
                        'cbm': 'Vasprun conduction band minimum',
                        'track': 'tracking for files',
                        'version': 'VASP executable version',
                        'nelect': 'NELECT' }

    def __init__( self, directory='.' ):
        self.directory = directory
        with cd( directory ):
            try:
                self.meta = VASPMeta.from_file('vaspmeta.yaml')
            except FileNotFoundError as e:
                raise FileNotFoundError(f'vaspmeta.yaml not found in {directory}')
            self.parse_vasprun()
        self.print_methods = { 'title': self.print_title,
                               'description': self.print_description,
                               'notes': self.print_notes,
                               'type': self.print_type,
                               'status': self.print_status,
                               'stoichiometry': self.print_stoichiometry,
                               'potcar': self.print_potcar,
                               'eatom': self.print_eatom,
                               'energy': self.print_energy,
                               'k-points': self.print_kpoints,
                               'functional': self.print_functional,
                               'encut': self.print_encut,
                               'plus_u': self.print_plus_u,
                               'ediffg': self.print_ediffg,
                               'ibrion': self.print_ibrion,
                               'converged': self.print_converged,
                               'version': self.print_version,
                               'md5': self.print_vasprun_md5,
                               'directory': self.print_directory,
                               'lreal': self.print_lreal,
                               'vbm': self.print_vbm,
                               'cbm': self.print_cbm,
                               'track': self.print_file_tracking, 
                               'nelect': self.print_nelect }
        if not set( self.print_methods.keys() ) == set( self.supported_flags ):
            print( set( self.print_methods.keys() ) )
            print( '--------------' )
            print( set( self.supported_flags.keys() ) )
            raise( ValueError )

    def parse_vasprun( self ):
        """
        Read in `vasprun.xml` as a pymatgen Vasprun object.

        Args:
            None

        Returns:
            None

        None:
            If the vasprun.xml is not well formed this method will catch the ParseError
            and set self.vasprun = None.
        """            
        self.vasprun_filename = match_filename( 'vasprun.xml' )
        if not self.vasprun_filename:
            raise FileNotFoundError( 'Could not find vasprun.xml or vasprun.xml.gz file' )
        try:
            self.vasprun = Vasprun( self.vasprun_filename, parse_potcar_file=False, 
                                    parse_dos=False )
        except ET.ParseError:
            self.vasprun = None
        except:
            raise

    @property
    def stoich( self ):
        return self.vasprun.final_structure.composition.get_el_amt_dict()

    @property
    def functional( self ):
        """
        String description of the calculation functional.
       
        Recognises:
            - PBE
            - PBEsol
            - PBE-based hybrids:
                - PBE0 (alpha=0.25, no screening)
                - HSE06 (alpha=0.25, mu=0.2)
                - generic hybrids (alpha=?, no screening)
                - generic screened hybrids (alpha=?, mu=?)
        
        Returns:
            (Str): String describing the calculation functional.

        """
        if self.potcars_are_pbe(): # PBE base funtional
            if 'LHFCALC' in self.vasprun.parameters:
                alpha = float( self.vasprun.parameters['AEXX'] )
            else:
                alpha = 0.0
            if 'HFSCREEN' in self.vasprun.parameters:
                mu = float( self.vasprun.parameters['HFSCREEN'] )
            else:
                mu = 0
            if alpha > 0:
                if mu > 0: # screened hybrid
                    if ( mu == 0.2 ) and ( alpha == 0.25 ):
                        f = 'HSE06'
                    else:
                        f = "screened hybrid. alpha={}, mu={}".format( alpha, mu )
                else: # unscreened hybrid
                    if alpha == 0.25:
                        f = 'PBE0'
                    else: 
                        f = "hybrid. alpha={}".format( alpha )
            else: # not hybrid. Plain PBE or some variant.
                pbe_list = { 'PS': 'PBEsol',
                             'PE': 'PBE',
                             '91': 'PW91',
                             'RP': 'rPBE',
                             'AM': 'AM05' }
                f = pbe_list[ self.vasprun.parameters['GGA'].upper() ]
        else:
            f = 'not recognised'    
        return f

    def potcars_are_pbe( self ):
        return all( 'PBE' in s for s in self.vasprun.potcar_symbols )

    def output( self, to_print ):
        if not self.vasprun:
            to_print = [ 'title', 'type', 'status' ]
        print( "---" )
        for p in to_print:
            self.print_methods[ p ]()
        print( '', flush=True )
       
    def print_type( self ):
        if self.meta.type:
            print( "type: {}".format( self.meta.type ) )
 
    def print_title( self ):
        print( "title: {}".format( self.meta.title ) )
      
    def print_description( self ):
        print( "description: {}".format( self.meta.description.strip() ) )
 
    def print_notes( self ):
        if self.meta.notes:
            print( "notes: {}".format( self.meta.notes.strip() ) )
        else:
            print( "notes: ~" )
 
    def print_status( self ):
        print( "status: {}".format( self.meta.status ) )

    def print_lreal( self ):
        print( "lreal: {}".format( self.vasprun.parameters['LREAL'] ) )

    def print_stoichiometry( self ):
        print( "stoichiometry:" )
        for element in self.stoich:
            print( "    - {}: {}".format( element, int( self.stoich[ element ] ) ) )

    def print_potcar( self ):
        print( "potcar:" )
        for e, p in zip( self.stoich, self.vasprun.potcar_symbols ):
            print( "    - {}: {}".format( e, p ) )
        
    def print_energy( self ):
        # if this gets more options, it might be a good idea to set the
        # appropriate method using a dictionary?
        # or we could subclass Summary --> NEB_Summary ?
        if not self.meta.type:
            print( "energy: {}".format( self.vasprun.final_energy ) )    
        elif self.meta.type == 'neb':
            self.print_neb_energy()
        else:
            raise ValueError( "VASPMeta type not supported: {}".format( self.meta.type ) )

    def print_neb_energy( self ):
        image_00_energy = final_energy_from_outcar( '00/OUTCAR' )
        print( "reference energy: {} eV".format( image_00_energy ) )
        neb = NEBAnalysis.from_dir( '.' )
        print( "neb image energies:" )
        for i, e in enumerate( neb.energies ):
            print( "    - {:02d}: {:10.6f} eV".format( i, e ) )

    def print_version( self ):
        version_string = vasp_version_from_outcar( '{}/OUTCAR'.format( self.directory ) ).split()[0]
        print( "version: {}".format( version_string ) )

    def print_eatom( self ):
        # This is one way to try to uniquely identify the POTCARs used, because the
        # potcar_symbol (e.g. `Ti_pv 07Sep2000`) is not sufficient.
        print( "eatom:" )
        for e, eatom in zip( self.stoich, potcar_eatom_list_from_outcar( '{}/OUTCAR'.format( self.directory ) ) ):
            print( "    - {}: {} eV".format( e, eatom ) )
        
    def print_kpoints( self ):
        print( "k-points:" )
        print( "    scheme: {}".format( self.vasprun.kpoints.style ) )
        print( "    grid: {}".format( " ".join( str( k ) for k in self.vasprun.kpoints.kpts[0] ) ) )

    def print_functional( self ):
        print( "functional: {}".format( self.functional ) )

    def print_ibrion( self ):
        print( "ibrion: {}".format( self.vasprun.incar['IBRION'] ) )

    def print_ediffg( self ):
        print( "ediffg: {}".format( self.vasprun.incar['EDIFFG'] ) )

    def print_encut( self ):
        if 'ENCUT' in self.vasprun.incar:
            print( "encut: {}".format( self.vasprun.incar['ENCUT'] ) )
        elif 'ENMAX' in self.vasprun.incar:
            print( "encut: {}".format( self.vasprun.incar['ENMAX'] ) )

    def print_converged( self ):
        print( "converged: {}".format( self.vasprun.converged ) )

    def print_vasprun_md5( self ):
        print( "vasprun md5: {}".format( file_md5( "{}/{}".format( self.directory, self.vasprun_filename ) ) ) )

    def print_file_tracking( self ):
        if self.meta.track:
            print( "file tracking:" )
            for f, new_filename in self.meta.track.items():
                print( "    {}:".format( f ) )
                if not new_filename:
                    new_filename = f
                print( "        filename: {}".format( new_filename ) )
                filename = match_filename( self.directory + f )
                if filename: 
                    md5 = file_md5( filename )
                else:
                    md5 = 'null'
                print( "        md5: {}".format( md5 ) )
 
    def print_directory( self ):
        print( "directory: {}".format( self.directory ) )

    def print_plus_u( self ):
        if 'LDAUU' in self.vasprun.incar:
            lqn = { 0: 's', 1: 'p', 2: 'd', 3: 'f' }
            ldauu = self.vasprun.incar[ 'LDAUU' ]
            ldauj = self.vasprun.incar[ 'LDAUJ' ]
            ldaul = self.vasprun.incar[ 'LDAUL' ]
            if any( v != 0 for v in ldauu ):
                print( 'ldau:' )
                for e, u, j, l in zip( self.stoich, ldauu, ldauj, ldaul ):
                    if u != 0:
                        print( "    - {}: {} {} {}".format( e, lqn[l], u, j ) )

    def print_cbm( self ):
        print( 'cbm: {}'.format( self.vasprun.eigenvalue_band_properties[1] ) )

    def print_vbm( self ):
        print( 'vbm: {}'.format( self.vasprun.eigenvalue_band_properties[2] ) )

    def print_nelect( self ):
        print( 'nelect: {}'.format( self.vasprun.parameters['NELECT'] ) )
        



#! /usr/bin/env python3

from pymatgen.io.vasp import Xdatcar  # type: ignore
from vasppy.rdf import RadialDistributionFunction
import argparse
import copy
import math
import numpy as np  # type: ignore

def parse_command_line_arguments():
    # command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument( 'xdatcar' )
    parser.add_argument( 'label', nargs=2 )
    parser.add_argument( 'max_r', type=float )
    parser.add_argument( 'n_bins', type=int )
    args = parser.parse_args()
    return( args )

def main():
    args = parse_command_line_arguments()
    max_r = args.max_r
    number_of_bins = args.n_bins
    species_1 = args.label[ 0 ]
    species_2 = args.label[ 1 ]
    xdatcar = Xdatcar( args.xdatcar )
    indices_i = [ i for i, s in enumerate(xdatcar.structures[0]) 
                  if s.species_string == species_1 ]
    if not indices_i:
        raise ValueError( f'No species {species_1} found' )
    indices_j = [ i for i, s in enumerate(xdatcar.structures[0]) 
                  if s.species_string == species_2 ]
    if not indices_j:
        raise ValueError( f'No species {species_2} found' )
    rdf = RadialDistributionFunction( xdatcar.structures, indices_i, indices_j, 
                                      number_of_bins, 0.0, max_r )
    for x, y in zip( rdf.r, rdf.rdf ):
        print(x, y)

if __name__ == "__main__":
    main()



#! /usr/bin/env python3

from vasppy import procar
from vasppy.outcar import reciprocal_lattice_from_outcar
import argparse

#def x_axis( cartesian_k_points ):
#    if cartesian_k_points is not None:
#        x_axis = [ 0.0 ]
#        for i in range( 1, len(cartesian_k_points) ):
#            d = cartesian_k_points[i - 1] - cartesian_k_points[i]
#            d = np.sqrt( np.dot( d, d) )
#            x_axis.append( d + x_axis[-1] )
#        x_axis = np.array( x_axis )
#    else:
#        x_axis = np.arange( len( cartesian_k_points ) ) 
#    return x_axis

def orbitals_with_l( l ):
    to_return = { 's' : [ 0 ],
                  'p' : [ 1, 2, 3 ],
                  'd' : [ 4, 5, 6, 7, 8 ],
                  'f' : [ 9, 10, 11, 12, 13 ],
                  'all' : None }
    return to_return[ l ]
 
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument( '-i', '--ions', help='ion indices for band projection (default: sum over all ions)', nargs='+', type=int )
    parser.add_argument( '-s', '--spins', help='spin indices for band projection (default [ 1 ])', nargs='+', type=int, default=[ 1 ] )
    parser.add_argument( '-o', '--orbitals', help='orbital indices for band projection (default: sum over all orbitals)', nargs='+', type=int )
    parser.add_argument( '-e', '--efermi', help='set fermi energy as reference for energy scale', type=float, default=0.0 )
    parser.add_argument( '-l', '--l-angular-momentum', help='select all orbitals with angular momentum L for band projection. This supercedes the --orbitals option', choices=[ 's', 'p', 'd', 'f', 'all' ] )
    parser.add_argument( '-f', '--procar', help='PROCAR filename (default PROCAR)', type=str, default='PROCAR' )
    parser.add_argument( '--scaling', help='Energy scaling for band widths (default 0.2 eV)', type=float, default=0.2 )
    parser.add_argument( '-x', '--xscaling', help='Automatic scaling of x-axis using reciprocal lattice vectors read from OUTCAR', action='store_true', default=False )
    args = parser.parse_args()

    if args.l_angular_momentum:
        args.orbitals = orbitals_with_l( args.l_angular_momentum )

    if args.xscaling:
        reciprocal_lattice = reciprocal_lattice_from_outcar( 'OUTCAR' ) # Move reading the reciprocal lattice to procar.py
    else:
        reciprocal_lattice = None

    pcar = procar.Procar()
    pcar.read_from_file( args.procar )
    pcar.print_weighted_band_structure( spins = args.spins, 
                                        ions = args.ions, 
                                        orbitals = args.orbitals, 
                                        scaling = args.scaling, 
                                        e_fermi = args.efermi, 
                                        reciprocal_lattice = reciprocal_lattice )

if __name__ == '__main__':
    main()



#! /usr/bin/env python3 

from vasppy.poscar import Poscar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer  # type: ignore
import argparse

def parse_command_line_arguments():
    # command line arguments
    parser = argparse.ArgumentParser( description='Finds the spacegroup for a VASP POSCAR file' )
    parser.add_argument( 'poscar', help="filename of the VASP POSCAR to be processed" )
    parser.add_argument( '-s', '--symprec', type=float, help="Precision for symmetry analuysis (defalut=1e-3)", default=1e-3 )
    args = parser.parse_args()
    return( args )

def main():
    args = parse_command_line_arguments()
    # initialise
    poscar = Poscar() # this doesn't really need vasppy. Could just use pymatgen to read the POSCAR
    # read POSCAR file
    poscar.read_from( args.poscar )
    structure = poscar.to_pymatgen_structure()
    symmetry_analyzer = SpacegroupAnalyzer( structure, symprec = args.symprec )
    print( symmetry_analyzer.get_space_group_symbol() )

if __name__ == "__main__":
    main()



import hashlib
from monty.io import zopen  # type: ignore
from pathlib import Path
import os
import yaml
from contextlib import contextmanager

@contextmanager
def cd( path ):
    old_dir = os.getcwd()
    os.chdir( path )
    try:
        yield
    finally:
        os.chdir( old_dir )

def md5sum( string ):
    """
    Generate the md5 checksum for a string

    Args:
        string (Str): The string to be checksummed.

    Returns:
        (Str): The hex checksum.
    """
    h = hashlib.new( 'md5' )
    h.update( string.encode( 'utf-8' ) )
    return h.hexdigest()

def file_md5( filename ):
    """
    Generate the md5 checksum for a file

    Args:
        filename (Str): The file to be checksummed.

    Returns:
        (Str): The hex checksum

    Notes:
        If the file is gzipped, the md5 checksum returned is
        for the uncompressed ASCII file.
    """
    with zopen( filename, 'r' ) as f:
        file_string = f.read()
    try: # attempt to decode byte object
        file_string = file_string.decode()
    except AttributeError:
        pass
    return( md5sum( file_string ) )

def match_filename( filename ):
    """
    Checks whether a file exists, either as named, or as a a gzippped file (filename.gz)

    Args:
        (Str): The root filename.

    Returns:
        (Str|None): if the file exists (either as the root filename, or gzipped), the return
            value will be the actual filename. If no matching filename is found the return
            value is set to None
    """
    f = next( ( '{}{}'.format( filename, extension ) for extension in [ '', '.gz' ]
        if Path( '{}{}'.format( filename, extension ) ).is_file() ), None ) 
    return f
   
def validate_checksum( filename, md5sum ):
    """
    Compares the md5 checksum of a file with an expected value.
    If the calculated and expected checksum values are not equal, 
    ValueError is raised.
    If the filename `foo` is not found, will try to read a gzipped file named
    `foo.gz`. In this case, the checksum is calculated for the unzipped file.

    Args:
        filename (str): Path for the file to be checksummed.
        md5sum (str):  The expected hex checksum.

    Returns:
        None
    """
    filename = match_filename( filename )
    md5_hash = file_md5( filename=filename )
    if md5_hash != md5sum:
        raise ValueError('md5 checksums are inconsistent: {}'.format( filename ))



#! /usr/bin/env python3

from vasppy.poscar import Poscar
import math
import argparse

def parse_command_line_arguments():
    parser = argparse.ArgumentParser( description='Rotates the cell lattice in VASP POSCAR files' )
    parser.add_argument( 'poscar', help="filename of the VASP POSCAR to be processed" )
    parser.add_argument( '-a', '--axis', nargs=3, type=float, help="vector for rotation axis", required=True )
    parser.add_argument( '-d', '--degrees', type=int, help="rotation angle in degrees", required=True )
    args = parser.parse_args()
    return( args )

def main():
    args = parse_command_line_arguments()
    poscar = Poscar()
    poscar.read_from( args.poscar )
    theta = math.pi * args.degrees / 180.0
    poscar.cell.rotate( args.axis, theta )
    poscar.output()

if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import numpy as np  # type: ignore
import argparse
from vasppy.poscar import Poscar
from vasppy.cell import Cell
from vasppy.pimaim import get_cart_coords_from_pimaim_restart

def parse_command_line_arguments():
    parser = argparse.ArgumentParser( description = 'TODO' )
    parser.add_argument( '-l', '--labels', nargs='+', help="labels for each species", required = True )
    parser.add_argument( '-n', '--atom-numbers', nargs='+', help="atom numbers for each species", required = True, type = int )
    args = parser.parse_args()
    if len( args.labels ) != len( args.atom_numbers ):
        print( '\'labels\' and \'atom-numbers\' require matched argument numbers' )
        exit()
    return args 

def lines_to_numpy_array( data ):
    return np.array( [ [ float( s ) for s in line.split() ] for line in data ] )

def read_pimaim_restart( filename ):
    with open( filename, 'r' ) as f:
        file_data = f.readlines()
    cr_dump_log, vel_dump_log, chg_dump_log, full_dump_log = [ bool( line.strip() ) for line in file_data[:4] ]
    # figure out how many atoms are in this calculation
    data_per_line = [ len( line.split() ) for line in file_data ] 
    number_of_atomic_data_lines = next( index for index, d in enumerate( data_per_line[4:] ) if d == 1 )
    number_of_atomic_data_types = sum( [ cr_dump_log, vel_dump_log, chg_dump_log ] )
    number_of_atoms = int( number_of_atomic_data_lines / number_of_atomic_data_types )
    # this assumes coordinates, velocities, and dipoles are all present.
    # not sure what happens if atoms have qudrupoles, etc.
    coordinates = lines_to_numpy_array( file_data[ 4 : 4 + number_of_atoms ] ) 
    velocities  = lines_to_numpy_array( file_data[ 4 + number_of_atoms : 4 + number_of_atoms * 2 ] ) 
    dipoles     = lines_to_numpy_array( file_data[ 4 + number_of_atoms * 2 : 4 + number_of_atoms * 3 ] ) 
    cell_matrix = lines_to_numpy_array( file_data[ -6: -3 ] )
    cell_lengths = lines_to_numpy_array( file_data[ -3: ] )
    full_cell_matrix = cell_matrix * cell_lengths
    return( coordinates, velocities, dipoles, full_cell_matrix, cell_lengths)

def main():
    filename = 'testout.rst'
    restart_file = True
    args = parse_command_line_arguments()
    coordinates, velocities, dipoles, full_cell_matrix, cell_lengths = read_pimaim_restart( filename )
    assert( sum( args.atom_numbers ) == len( coordinates ) )
    poscar = Poscar()
    full_cell_matrix = full_cell_matrix.transpose()
    coordinates = get_cart_coords_from_pimaim_restart(coordinates, full_cell_matrix,cell_lengths)
    poscar.cell = Cell( full_cell_matrix)
    poscar.atoms = args.labels
    poscar.atom_numbers = args.atom_numbers
    poscar.coordinate_type = 'Cartesian'
    poscar.coordinates = coordinates
    poscar.output()

if __name__ == '__main__':
    main()



from lxml import etree # type: ignore
from typing import List, Union, Optional, Any, Dict
from pymatgen import Structure # type: ignore
import numpy as np # type: ignore

def parse_varray(varray: etree.Element) -> Union[List[List[float]], 
                                                 List[List[int]],
                                                 List[List[bool]]]:
    """Parse <varray> data.

    Args:
        varray (etree.Element): xml <varray> element.

    Returns:
        (list(list): A nested list of either float, int, or bool.

    """
    m: Union[List[List[int]], List[List[float]], List[List[bool]]]
    varray_type = varray.get("type", None)
    v_list = [v.text.split() for v in varray.findall("v")] 
    if varray_type == 'int':
        m = [[int(number) for number in v] for v in v_list]
    elif varray_type == 'logical':
        m = [[i == "T" for i in v] for v in v_list]
    else:
        m = [[float(number) for number in v] for v in v_list]
    return m

def parse_structure(structure: etree.Element) -> Dict[str, Any]:
    """Parse <structure> data..

    Args:
        structure (etree.Element): xml <structure> element.

    Returns:
        (dict): Dictionary of structure data:
            `lattice`: cell matrix (list(list(float))..
            `frac_coords`: atom fractional coordinates (list(list(float)).
            `selective_dynamics`: selective dynamics (list(bool)|None).

    """
    latt = parse_varray(structure.find("crystal").find("varray"))
    pos = parse_varray(structure.find("varray"))
    sdyn = structure.find("varray/[@name='selective']")
    if sdyn:
        sdyn = parse_varray(sdyn)
    structure_dict = {'lattice': latt,
                      'frac_coords': pos,
                      'selective_dynamics': sdyn}
    return structure_dict

def structure_from_structure_data(lattice: List[List[float]],
                                  atom_names: List[str],
                                  frac_coords: List[List[float]]) -> Structure:
    """Generate a pymatgen Structure.

    Args:
        lattice (list(list(float)): 3x3 cell matrix.
        atom_names (list(str)): list of atom name strings.
        frac_coords (list(list(float): Nx3 list of fractional coordinates.

    Returns:
        (pymatgen.Structure)

    """
    structure = Structure(lattice=lattice,
                          species=atom_names,
                          coords=frac_coords,
                          coords_are_cartesian=False)
    return structure

class Vasprun:
    """Object for parsing vasprun.xml data.

    Attributes:
        atom_names (list(str)): List of atom name strings.
        structures (list(pymatgen.Structure): List of structures as pymatgen Structure objects.
        frac_coords (np.array): timesteps x atoms x 3 numpy array of fractional coordinates.
        cart_coords (np.array): timesteps x atoms x 3 numpy array of cartesian coordinates.
        forces (:obj:`np.array`, optional): timesteps x atoms x 3 numpy array of forces.

    Examples:
        >>> vasprun = Vasprun('vasprun.xml')
        >>> cart_coords = vasprun.cart_coords
        >>> forces = vasprun.forces

    """

    def __init__(self,
                 filename: str) -> None: 
        """Initialise a Vasprun object from a vasprun.xml file.

        Args:
            filename (str): The vasprun.xml filename.

        Returns:
            None

        """
        doc = etree.parse(filename)
        self.doc = doc.getroot()
        self._atom_names = None # type: Optional[List[str]]
        self._structures = None # type: Optional[List[Structure]]

    @property
    def structures(self) -> List[Structure]:
        """Getter for structures attribut.

        Returns:
            (list(pymatgen.Structure)): A list of pymatgen Structure objects.

        Notes:
            When first called this parses the vasprun XML data and
            caches the result.

        """
        if not self._structures:
            self._structures = self.parse_structures()
        return self._structures
 
    @property
    def atom_names(self) -> List[str]:
        """Getter for atom_names attribute.

        Returns:
            (list(str)): A list of atom name strings.

        Notes:
            When first called this parses the vasprun XML data and
            caches the result.

        """
        if not self._atom_names:
            self._atom_names = self.parse_atom_names()
        return self._atom_names

    def parse_atom_names(self) -> List[str]:
        """Return a list of atom names for the atoms in this calculation.

        Args:
            None

        Returns:
            (list(str))

        """
        atominfo = self.doc.find("atominfo")
        if atominfo is None:
            raise ValueError("No atominfo found in file")
        atom_names = []
        for array in atominfo.findall("array"):
            if array.attrib["name"] == "atoms":
                atom_names = [rc.find("c").text.strip() for rc in array.find("set")]
        if not atom_names:
            raise ValueError("No atomname found in file")
        return atom_names 

    def parse_structures(self) -> List[Structure]:
        """Returns a list of pymatgen Structures for this calculation.

        Args:
            None

        Returns:
            (list(pymatgen.Structure))

        """
        structures = []
        for child in self.doc.iterfind("calculation"):
            elem = child.find("structure")
            structure_data = parse_structure(elem)
            structures.append(
                structure_from_structure_data(
                    lattice=structure_data['lattice'],
                    atom_names=self.atom_names,
                    frac_coords=structure_data['frac_coords']
                )
            )
        return structures

    @property
    def frac_coords(self) -> np.ndarray:
        """Fractional coordinates from each calculation structure.

        Args:
            None

        Returns:
            (np.ndarray): timesteps x atoms x 3 numpy array of fractional coordinates.

        """
        frac_coords = np.array([s.frac_coords for s in self.structures])
        return frac_coords

    @property
    def cart_coords(self) -> np.ndarray:
        """Cartesian coordinates from each calculation structure.

        Args:
            None

        Returns:
            (np.ndarray): timesteps x atoms x 3 numpy array of cartesian coordinates.

        """
        cart_coords = np.array([s.cart_coords for s in self.structures])
        return cart_coords

    @property
    def forces(self) -> Optional[np.ndarray]:
        """Cartesian forces from each calculation structure
           (if present in the vasprun XML).

        Args:
            None

        Returns:
            (np.ndarray|None): timesteps x atoms x 3 numpy array of cartesian forces
                if forces are included in the vasprun XML. If not, returns None. 

        """
        forces = []
        for child in self.doc.iterfind("calculation"):
            elem = child.find("varray/[@name='forces']")
            if elem != None:
                forces.append(parse_varray(elem))
        if forces:
            return np.array(forces)
        else:
            return None
 



import numpy as np  # type: ignore
import copy

class Polyhedron:

    def __init__( self, vertices, cell, inside_point, cutoff ):
        self.vertices = vertices
        self.cell = cell
        self.inside_point = inside_point
        self.vertices = [ self.cell.nearest_image( inside_point, p ) for p in vertices ]
        self.inside_point = self.centre() # improved inside_point

    def centre( self ):
        return( sum( self.vertices ) / len( self.vertices ) )

    def print_points( self ):
        for point in self.vertices:
            print( point.dot( self.cell.matrix ) )



import numpy as np  # type: ignore
import pandas as pd  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
from matplotlib.axis import Axes  # type: ignore
from matplotlib.figure import Figure  # type: ignore
import matplotlib._color_data as mcd  # type: ignore
from typing import Optional, List, Union, Dict, Tuple
from collections.abc import Iterable 


TABLEAU_GREY: str = '#bab0ac'


def pdos_column_names(lmax: int, ispin: int) -> List[str]:
    if lmax == 2:
        names = ['s', 'p_y', 'p_z', 'p_x', 'd_xy',
                 'd_yz', 'd_z2-r2', 'd_xz', 'd_x2-y2']
    elif lmax == 3:
        names = ['s',
                 'p_y', 'p_z', 'p_x',
                 'd_xy', 'd_yz', 'd_z2-r2', 'd_xz', 'd_x2-y2',
                 'f_y(3x2-y2)', 'f_xyz', 'f_yz2', 'f_z3',
                 'f_xz2', 'f_z(x2-y2)', 'f_x(x2-3y2)']
    else:
        raise ValueError('lmax value not supported')
    if ispin == 2:
        all_names = []
        for n in names:
            all_names.extend(['{}_up'.format(n), '{}_down'.format(n)])
    else:
        all_names = names
    all_names.insert(0, 'energy')
    return all_names


class Doscar:
    '''Contains all the data in a VASP DOSCAR file,
    and methods for manipulating this.
    '''

    number_of_header_lines: int = 6

    def __init__(self,
                 filename: str,
                 ispin: int = 2,
                 lmax: int = 2,
                 lorbit: int = 11,
                 spin_orbit_coupling: bool = False,
                 read_pdos: bool = True, 
                 species: Optional[List[str]] = None) -> None:
        '''
        Create a Doscar object from a VASP DOSCAR file.

        Args:
            filename (str): Filename of the VASP DOSCAR file to read.
            ispin (optional:int): ISPIN flag.
                Set to 1 for non-spin-polarised 
                or to 2 for spin-polarised calculations.
                Default = 2.
            lmax (optional:int): Maximum l angular momentum. (d=2, f=3). Default is 2.
            lorbit (optional:int): The VASP LORBIT flag. (Default=11).
            spin_orbit_coupling (optional:bool): Spin-orbit coupling (Default=False).
            read_pdos (optional:bool): Set to True to read the atom-projected density of states (Default=True).
            species (optional:list(str)): List of atomic species strings, e.g. `['Fe', 'Fe', 'O', 'O', 'O']`.
                Default=None.
        '''
        self.filename = filename
        self.ispin = ispin
        self.lmax = lmax
        self.spin_orbit_coupling = spin_orbit_coupling
        if self.spin_orbit_coupling:
            raise NotImplementedError(
                'Spin-orbit coupling is not yet implemented')
        self.lorbit = lorbit
        self.pdos: Optional[np.array] = None
        self.species = species
        self.read_header()
        self.read_total_dos()
        if read_pdos:
            try:
                self.read_projected_dos()
            except:
                raise
        # if species is set, should check that this is consistent with the number of entries in the
        # projected_dos dataset

    @property
    def number_of_channels(self) -> int:
        if self.lorbit == 11:
            return {2: 9, 3: 16}[self.lmax]
        raise NotImplementedError

    def read_header(self) -> None:
        self.header = []
        with open(self.filename, 'r') as file_in:
            for i in range(Doscar.number_of_header_lines):
                self.header.append(file_in.readline())
        self.process_header()

    def process_header(self) -> None:
        self.number_of_atoms = int(self.header[0].split()[0])
        self.number_of_data_points = int(self.header[5].split()[2])
        self.efermi = float(self.header[5].split()[3])

    def read_total_dos(self) -> pd.DataFrame:  # assumes spin_polarised
        start_to_read: int = Doscar.number_of_header_lines
        df: pd.DataFrame = pd.read_csv(self.filename,
                                       skiprows=start_to_read,
                                       nrows=self.number_of_data_points,
                                       delim_whitespace=True,
                                       names=['energy', 'up', 'down', 'int_up', 'int_down'],
                                       index_col=False)
        self.energy: np.ndarray = df.energy.values
        df.drop('energy', axis=1)
        self.tdos = df

    # currently assume spin-polarised, no-SO-coupling, no f-states
    def read_atomic_dos_as_df(self, atom_number: int) -> pd.DataFrame:
        assert atom_number > 0 & atom_number <= self.number_of_atoms
        start_to_read = Doscar.number_of_header_lines + \
            atom_number * (self.number_of_data_points + 1)
        df = pd.read_csv(self.filename,
                         skiprows=start_to_read,
                         nrows=self.number_of_data_points,
                         delim_whitespace=True,
                         names=pdos_column_names(
                             lmax=self.lmax, ispin=self.ispin),
                         index_col=False)
        return df.drop('energy', axis=1)

    def read_projected_dos(self) -> None:
        """Read the projected density of states data"""
        pdos_list = []
        for i in range(self.number_of_atoms):
            df = self.read_atomic_dos_as_df(i + 1)
            pdos_list.append(df)
        self.pdos = np.vstack([np.array(df) for df in pdos_list]).reshape(
            self.number_of_atoms, self.number_of_data_points, self.number_of_channels, self.ispin)

    def pdos_select(self, 
                    atoms: Optional[Union[int, List[int]]] = None,
                    spin: Optional[str] = None,
                    l: Optional[str] = None,
                    m: Optional[List[str]] = None) -> np.ndarray:
        """
        Returns a subset of the projected density of states array.

        Args:
            atoms (int or list(int)): Atom numbers to include in the selection. Atom numbers count from 1. 
                                   Default is to select all atoms.
            spin (str): Select up or down, or both spin channels to include in the selection.
                        Accepted options are 'up', 'down', and 'both'. Default is to select both spins.
            l (str): Select one angular momentum to include in the selectrion.
                     Accepted options are 's', 'p', 'd', and 'f'. Default is to include all l-values.
                     Setting `l` and not setting `m` will return all projections for that angular momentum value.
            m (list(str)): Select one or more m-values. Requires `l` to be set. 
                           The accepted values depend on the value of `l`:
                           `l='s'`: Only one projection. Not set.
                           `l='p'`: One or more of ['x', 'y', 'z']
                           `l='d'`: One or more of ['xy', 'yz', 'z2-r2', 'xz', 'x2-y2']
                           `l='f'`: One or more of ['y(3x2-y2)', 'xyz', 'yz2', 'z3', 'xz2', 'z(x2-y2)', 'x(x2-3y2)']

        Returns:
            np.array: A 4-dimensional numpy array containing the selected pdos values. 
            The array dimensions are [ atom_no, energy_value, lm-projection, spin ]

        """
        assert isinstance(self.pdos, np.ndarry)
        valid_m_values: Dict[str, List[str]] = {'s': [],
                          'p': ['x', 'y', 'z'],
                          'd': ['xy', 'yz', 'z2-r2', 'xz', 'x2-y2'],
                          'f': ['y(3x2-y2)', 'xyz', 'yz2', 'z3', 'xz2', 'z(x2-y2)', 'x(x2-3y2)']}
        if not atoms:
            atom_idx = list(range(self.number_of_atoms))
        else:
            assert isinstance(atoms, list)
            atom_idx = atoms
        to_return = self.pdos[atom_idx, :, :, :]
        if not spin:
            spin_idx = list(range(self.ispin))
        elif spin is 'up':
            spin_idx = [0]
        elif spin is 'down':
            spin_idx = [1]
        elif spin is 'both':
            spin_idx = [0, 1]
        else:
            raise ValueError(
                "valid spin values are 'up', 'down', and 'both'. The default is 'both'")
        to_return = to_return[:, :, :, spin_idx]
        if not l:
            channel_idx = list(range(self.number_of_channels))
        elif l == 's':
            channel_idx = [0]
        elif l == 'p':
            if not m:
                channel_idx = [1, 2, 3]
            else:  # TODO this looks like it should be i+1
                channel_idx = [i + 1 for i, v in enumerate(valid_m_values['p']) if v in m]
        elif l == 'd':
            if not m:
                channel_idx = [4, 5, 6, 7, 8]
            else:  # TODO this looks like it should be i+4
                channel_idx = [i + 4 for i, v in enumerate(valid_m_values['d']) if v in m]
        elif l == 'f':
            if not m:
                channel_idx = [9, 10, 11, 12, 13, 14, 15]
            else:  # TODO this looks like it should be i+9
                channel_idx = [i + 9 for i, v in enumerate(valid_m_values['f']) if v in m]
        else:
            raise ValueError
        return to_return[:, :, channel_idx, :]

    def pdos_sum(self,                    
                 atoms: Optional[Union[int, List[int]]] = None,
                 spin: Optional[str] = None,
                 l: Optional[str] = None,
                 m: Optional[List[str]] = None) -> np.ndarray: 
        return np.sum(self.pdos_select(atoms=atoms, spin=spin, l=l, m=m), axis=(0, 2, 3))

    def plot_pdos(self, 
                  ax: Optional[Axes] = None,
                  to_plot: Optional[Dict[str, List[str]]] = None,
                  colors: Optional[Iterable] = None,
                  plot_total_dos: Optional[bool] = True,
                  xrange: Optional[Tuple[float, float]] = None, 
                  ymax: Optional[float] = None,
                  scaling: Optional[Dict[str, Dict[str, float]]] = None,
                  split: bool = False,
                  title: Optional[str] = None,
                  title_loc: str = 'center',
                  labels: bool = True,
                  title_fontsize: int = 16,
                  legend_pos: str = 'outside') -> Figure:
        if not ax:
            fig, ax = plt.subplots(1, 1, figsize=(8.0, 3.0))
        else:
            fig = None
        if not colors:
            colors = mcd.TABLEAU_COLORS
        assert isinstance(colors, Iterable)
        color_iterator = (c for c in colors)

        if not scaling:
            scaling = {}

        if xrange:
            e_range = (self.energy >= xrange[0]) & (self.energy <= xrange[1])
        else:
            e_range = np.ma.make_mask(self.energy)

        auto_ymax = 0.0

        if not to_plot:
            to_plot = {}
            assert isinstance(self.species, Iterable)
            for s in set(self.species):
                to_plot[s] = ['s', 'p', 'd']
                if self.lmax == 3:
                    to_plot[s].append('f')

        for species in to_plot.keys():
            assert isinstance(self.species, Iterable)
            index = [i for i, s in enumerate(self.species) if s is species]
            for state in to_plot[species]:
                assert state in ['s', 'p', 'd', 'f']
                color = next(color_iterator)
                label = '{} {}'.format(species, state)
                up_dos = self.pdos_sum(atoms=index, l=state, spin='up')[e_range]
                down_dos = self.pdos_sum(atoms=index, l=state, spin='down')[e_range]
                if species in scaling:
                    if state in scaling[species]:
                        up_dos *= scaling[species][state]
                        down_dos *= scaling[species][state]
                        label = r'{} {} $\times${}'.format(
                            species, state, scaling[species][state])
                auto_ymax = max([auto_ymax, up_dos.max(), down_dos.max()])
                ax.plot(self.energy[e_range], up_dos, label=label, c=color)
                ax.plot(self.energy[e_range], down_dos * -1.0, c=color)
        if plot_total_dos:
            ax.fill_between(self.energy[e_range], self.tdos.up.values[e_range],
                            self.tdos.down.values[e_range] * -1.0, facecolor=TABLEAU_GREY, alpha=0.2)
            auto_ymax = max([auto_ymax, self.tdos.up.values[e_range].max(
            ), self.tdos.down.values[e_range].max()])

        if xrange:
            ax.set_xlim(xrange[0], xrange[1])

        if not ymax:
            ymax = 1.1 * auto_ymax
        ax.set_ylim(-ymax * 1.1, ymax * 1.1)
        if legend_pos == 'outside':
            ax.legend(bbox_to_anchor=(1.01, 1.04), loc='upper left')
        else:
            ax.legend(loc=legend_pos)
        if labels:
            ax.set_xlabel('Energy [eV]')
        ax.axhline(y=0, c='lightgrey')
        ax.axes.grid(False, axis='y')

        ax.tick_params(
            axis='y',        # changes apply to the y-axis
            which='both',    # both major and minor ticks are affected
            left=False,      # ticks along the left edge are off
            right=False,     # ticks along the right edge are off
            labelleft=False)  # labels along the left edge are off

        if title:
            ax.set_title(title, loc=title_loc, fontdict={
                         'fontsize': title_fontsize})

        return fig



#! /usr/bin/env python3

from vasppy.poscar import Poscar
from vasppy.summary import potcar_spec, potcar_sets
import argparse
import re

"""
A command line utility for testing species consistency between a VASP POSCAR and POTCAR pair of files. Species are considered consistent if the species labels used in the POSCAR file match the start of the pseudopotential labels in the POTCAR file, in order. e.g. a POSCAR that contains `Ti O` will match a POTCAR that contains `Ti_pv O`. If any species labels do not match the script raises an AttributeError.

The `-p` flag will check that all the pseudopotentials in the POTCAR file belong to a specific pseudopotential set.
"""

def parse_command_line_arguments():
    parser = argparse.ArgumentParser( description='Check species consistency between a VASP POSCAR file and a POTCAR file.' )
    parser.add_argument( 'poscar', help="filename of the VASP POSCAR to be processed", nargs='?', default='POSCAR' )
    parser.add_argument( 'potcar', help="filename of the VASP POTCAR to be processed", nargs='?', default='POTCAR' )
    parser.add_argument( '-p', '--ppset', help="check whether the POTCAR pseudopotentials belong to a specific pseudopotential set", choices=potcar_sets )
    return parser.parse_args()

def main():
    args = parse_command_line_arguments()
    poscar = Poscar.from_file( args.poscar )
    potcars = potcar_spec( args.potcar )
    for i, ( species, potcar ) in enumerate( zip( poscar.atoms, potcars ), 1 ):
        matching_potcar = potcar.startswith( species )
        if not matching_potcar:
            raise AttributeError( 'Species {} mismatch:\nPOSCAR contains {}\nPOTCAR contains {}'.format( i, species, potcar ) )
        if args.ppset:
            this_ppset = potcars[potcar]
            if args.ppset != this_ppset:
                raise AttributeError( 'Pseudopotential set mismatch: {}'.format( potcars.values() ) )

if __name__ == '__main__':
    main()



import yaml
from vasppy.utils import file_md5

class VASPMeta:
    """
    VASPMeta class for storing additional VASP calculation metadata
    """

    def __init__( self, title, description, status, notes=None, type=None, track=None ):
        """
        Initialise a VASPMeta object.

        Args:
            title (Str): The title string for this calculation
            description (Str): Long description
            status (Str): Current status of the calculation. 
                Expected strings are (to-run, incomplete, finished, dropped)
            notes (:obj:Str, optional): Any additional notes. Defaults to None.
            type  (:obj:Str, optional): Can be used to describe the calculation type. 
                Defaults to None.
            track (:obj:dict(str: str), optional): An optional dict of pairs of filenames
                for files to track. For each key: value pair, the key is the current filename
                in the directory. The value is the new filename that list of filenames to calculate md5 hashes
                files to calculate hashes for when summarising the calculation output.
                Defaults to None.

        Returns:
            None
        """
        self.title = title
        self.description = description
        self.notes = notes
        expected_status = [ 'to-run', 'incomplete', 'finished', 'dropped' ]
        if status not in expected_status:
            raise ValueError(f'Unexpected calculations status: "{status}"'
                             f' for calculation {title}')
        self.status = status
        expected_types = [ 'single-point', 'neb' ]
        if type:
            if type not in expected_types:
                raise ValueError(f'Unexpected calculation type: "{type}"'
                                 f' for calculation {title}')
            self.type = type 
        else:
            self.type = None
        self.track = track

    @classmethod
    def from_file( cls, filename ):
        """
        Create a VASPMeta object by reading a `vaspmeta.yaml` file

        Args:
            filename (Str): filename to read in.

        Returns:
            (vasppy.VASPMeta): the VASPMeta object
        """
        with open( filename, 'r' ) as stream:
            data = yaml.load( stream, Loader=yaml.SafeLoader )
            notes = data.get( 'notes' )
            v_type = data.get( 'type' )
            track = data.get( 'track' )
            xargs = {}
            if track:
                if type( track ) is str:
                    track = [ track ]
                xargs['track'] = track 
            vaspmeta = VASPMeta( data['title'], 
                                 data['description'], 
                                 data['status'], 
                                 notes=notes, 
                                 type=v_type,
                                 **xargs )
        return vaspmeta




import numpy as np  # type: ignore
import sys
import re
import copy
from vasppy import configuration, atom, cell
from vasppy.units import angstrom_to_bohr
from pymatgen import Lattice as pmg_Lattice  # type: ignore
from pymatgen import Structure as pmg_Structure  # type: ignore
from pymatgen.io.cif import CifWriter  # type: ignore
from collections import Counter

# Ignore SIG_PIPE and don't throw exceptions on it... 
# http://newbebweb.blogspot.co.uk/2012/02/python-head-ioerror-errno-32-broken.html
from signal import signal, SIGPIPE, SIG_DFL
signal( SIGPIPE, SIG_DFL ) 

def parity( list ):
    return( sum( list )%2 )

def swap_axes( matrix, axes ):
    axes_index = { 'x': 0, 'y': 1, 'z': 2 }
    matrix[:, [ axes_index[ axes[ 0 ] ], axes_index[ axes[ 1 ] ] ] ] = matrix[:, [ axes_index[ axes[ 1 ] ], axes_index[ axes[ 0 ] ] ] ]
    return matrix

class Poscar:

    lines_offset = 9

    def __init__( self ):
        self.title = "Title"
        self.scaling = 1.0
        self.cell = cell.Cell( np.identity( 3 ) )
        self.atoms = [ 'A' ]
        self.atom_numbers = [ 1 ]
        self.coordinate_type = 'Direct'
        self.coordinates = np.array( [ [ 0.0, 0.0, 0.0 ] ] )
        self.selective_dynamics = False

    def coordinates_by_species( self, species ):
        return self.coordinates[ self.range_by_species( species ) ]
 
    def range_by_species( self, species ):
        i = 0
        atom_range = {}
        for a, n in zip( self.atoms, self.atom_numbers ):
            atom_range[ a ] = range(i, i+n)
            i += n
        return atom_range[ species ]

    def atom_number_by_species( self, species ):
        atom_numbers = { a : n for a, n in zip( self.atoms, self.atom_numbers ) }
        return atom_numbers[ species ]

    def sorted( self, species ):
        new_poscar = copy.deepcopy( self )
        new_poscar.atoms = species
        new_poscar.atom_numbers = [ self.atom_number_by_species( s ) for s in species ]
        new_poscar.coordinates = np.concatenate( [ self.coordinates_by_species( s ) for s in species ], axis = 0 )
        return new_poscar 
 
    def read_from( self, filename ):
        with open( filename ) as f:
            lines = f.readlines()
        self.title = lines.pop(0).strip()
        self.scaling = float( lines.pop(0).strip() ) 
        self.cell.matrix = np.array( [ [ float( e ) for e in lines.pop(0).split() ] for i in range( 3 ) ] )
        self.cell.inv_matrix = np.linalg.inv( self.cell.matrix )
        self.atoms = lines.pop(0).split()
        self.atom_numbers = [ int(element) for element in lines.pop(0).split() ]
        self.coordinate_type = lines.pop(0).strip()
        if re.match( r'\A[Ss]', self.coordinate_type ): # test for 'Selective dynamics'
            self.selective_dynamics = True
            self.coordinate_type = lines.pop( 0 ).strip()
        self.coordinates = np.array( [ [ float( e ) for e in lines.pop(0).split()[0:3] ] for i in range( sum( self.atom_numbers ) ) ] )
        if self.coords_are_cartesian(): # Convert to direct coordinates
            self.coordinates = self.fractional_coordinates()
            self.coordinate_type = 'Direct'

    @classmethod
    def from_file( cls, filename ):
        poscar = cls()
        poscar.read_from( filename )
        return poscar

    def in_bohr( self ):
        new_poscar = copy.deepcopy( self )
        bohr_to_angstrom = 0.529177211
        new_poscar.scaling *= bohr_to_angstrom
        new_poscar.cell.matrix /= bohr_to_angstrom
        if new_poscar.coords_are_cartesian():
            new_poscar.coordinates /= bohr_to_angstrom
        return( new_poscar )

    def coords_are_fractional( self ): 
        return not self.coords_are_cartesian()

    def coords_are_cartesian( self ):
        return re.match( r'\A[CcKk]', self.coordinate_type )

    def fractional_coordinates( self ):
        return ( self.coordinates if self.coords_are_fractional() else self.coordinates.dot( np.linalg.inv( self.cell.matrix ) ) )
        # return ( self.coordinates if self.coords_are_fractional() else self.cell.cartesian_to_fractional_coordinates( coordinates )
            # need to check whether this alternative version works.

    def cartesian_coordinates( self ):
        return ( self.coordinates if self.coords_are_cartesian() else self.coordinates.dot( self.cell.matrix ) )

    def select_coordinates( self, coordinate_type='Direct' ):
        coord_opts = { 'Direct'    : self.fractional_coordinates(), 
                       'Cartesian' : self.cartesian_coordinates() }
        return coord_opts[ coordinate_type ]

    def output_coordinates_only( self, coordinate_type='Direct', opts=None ):
        prefix = []
        suffix = []
        for i in range( self.coordinates.shape[0] ):
            prefix_string = ''
            suffix_string = ''
            if 'selective' in opts:
                if opts['selective'] == 'T':
                    suffix_string += ' T T T'
                elif opts['selective'] == 'F':
                    suffix_string += ' F F F'
                elif opts['selective']:
                    raise ValueError
            if 'numbered' in opts:
                if opts['numbered']:
                    suffix_string += ' {}'.format(i+1)
            if 'label' in opts:
                if opts['label']:
                    if opts['label'] == 1:
                        prefix_string += self.labels()[i].ljust(6) 
                    elif opts['label'] == 4:
                        suffix_string += ' {}'.format(self.labels()[i])
                    elif opts['label']:
                        raise ValueError( opts['label'] )
            prefix.append( prefix_string )
            suffix.append( suffix_string )
        for pref, coord, suff in zip( prefix, self.select_coordinates( coordinate_type ), suffix ):
            print( pref + ''.join( ['  {: .10f}'.format( element ) for element in coord ] ) + suff )
  
    def output( self, coordinate_type='Direct', opts=None ):
        if opts is None:
            opts = {}
        if not opts.get( 'coordinates_only' ):
            self.output_header( coordinate_type=coordinate_type )
        self.output_coordinates_only( coordinate_type=coordinate_type, opts=opts )

    def output_header( self, coordinate_type='Direct', opts=None ):
        if opts is None:
            opts = {}
        print( self.title )
        print( self.scaling )
        for row in self.cell.matrix:
            print( ''.join( ['   {: .10f}'.format( element ) for element in row ] ) )
        print( ' '.join( self.atoms ) )
        print( ' '.join( [ str(n) for n in self.atom_numbers ] ) )
        if opts.get('selective'):
            print( 'Selective Dynamics' )
        print( coordinate_type )

    def write_to( self, filename, coordinate_type='Direct', opts=None ):
        if opts is None:
            opts = {}
        with open( filename, 'w' ) as sys.stdout:
            self.output( coordinate_type=coordinate_type, opts=opts )
        sys.stdout = sys.__stdout__ # make sure sys.stdout is reset
        
    def output_as_xtl( self ):
        print( self.title )
        print( "CELL" )
        cell_lengths = self.cell_lengths()
        cell_angles  = self.cell_angles()
        cell_data = cell_lengths + cell_angles
        print( ''.join( ['   {: .8f}'.format( element ) for element in cell_data ] ) )
        print( " Symmetry label P1\n\nATOMS\nNAME      X       Y     Z" )
        output_opts = { 'label' : True }
        self.output_coordinates_only( coordinate_type='Direct', opts = output_opts )

    def output_as_cif( self, symprec = None ):
        print( CifWriter( self.to_pymatgen_structure(), symprec ) )

    def output_as_pimaim( self, to_bohr = True ):
        if to_bohr is True:
            unit_scaling = angstrom_to_bohr
        else:
            unit_scaling = 1.0
        cell_lengths = self.cell.lengths() * self.scaling / unit_scaling
        print( "T\nF\nF\nF" )
        for row in self.cell_coordinates():
            print( ' '.join( [ str( x / unit_scaling ) for x in row ] ) )
        for row in self.cell.unit_vectors().transpose():
            print( ' '.join( [ str( x ) for x in row ] ) )
        for length in cell_lengths:
            print( length )

    def cell_coordinates( self ):
        return( self.coordinates * self.cell.lengths() * self.scaling )

    def labels( self ):
        return( [ atom_name for ( atom_name, atom_number ) in zip( self.atoms, self.atom_numbers ) for __ in range( atom_number ) ] )

    def replicate( self, h, k, l, group=False ):
        lattice_scaling = np.array( [ h, k, l ], dtype=float )
        lattice_shift = np.reciprocal( lattice_scaling ) 
        new_poscar = Poscar()
        new_poscar.title = self.title
        new_poscar.scaling = self.scaling
        # print( lattice_scaling.dot( self.lattice ) )
        new_poscar.cell.matrix = ( self.cell.matrix.T * lattice_scaling ).T
        new_poscar.coordinate_type = self.coordinate_type
        new_coordinate_list = []
        cell_shift_indices = [ [ a,b,c ] for a in range(h) for b in range(k) for c in range(l) ]
        # generate grouped / ungrouped atom names and numbers for supercell
        if group:
            new_poscar.atoms = [ label + group for label in self.atoms for group in ('a','b') ]
            new_poscar.atom_numbers = [ int( num * h * k * l / 2 ) for num in self.atom_numbers for __ in ( 0, 1 )]
        else:
            new_poscar.atoms = self.atoms
            new_poscar.atom_numbers = [ num * h * k * l for num in self.atom_numbers ]
        # generate grouped / ungrouped atoms coordinates for supercell
        if group:
            for row in self.coordinates:
                pos_in_origin_cell = row / lattice_scaling
                for odd_even in ( 0, 1 ):
                    for ( a, b, c ) in [ cell_shift for cell_shift in cell_shift_indices if parity(cell_shift) == odd_even ]:
                        new_coordinate_list.append( [ pos_in_origin_cell + np.array( [ a, b, c ] * lattice_shift ) ][0].tolist() )
        else:
            for row in self.coordinates:
                pos_in_origin_cell = row / lattice_scaling
                for ( a, b, c ) in cell_shift_indices:
                    new_coordinate_list.append( [ pos_in_origin_cell + np.array( [ a, b, c ] * lattice_shift ) ][0].tolist() )
        new_poscar.coordinates = np.array( new_coordinate_list )
        return new_poscar    

    def cell_lengths( self ):
        return( ( self.cell.lengths() * self.scaling ).tolist() )
        # return [ np.linalg.norm( row ) for row in self.cell.matrix * self.scaling ]

    def cell_angles( self ):
        return( self.cell.angles() )
        # ( a, b, c ) = [ row for row in self.cell.matrix ]
        # return [ angle( b, c ), angle( a, c ), angle( a, b ) ]

    def to_configuration( self ):
        atoms = [ atom.Atom( label, coordinates ) for ( label, coordinates ) in zip( self.labels(), self.fractional_coordinates() ) ]
        config = configuration.Configuration( cell.Cell( matrix = self.cell.matrix * self.scaling ), atoms )
        return( config )

    def swap_axes( self, axes ):
        new_poscar = copy.deepcopy( self )
        new_poscar.cell = cell.Cell( swap_axes( self.cell.matrix, axes ) )
        return new_poscar

    def to_pymatgen_structure( self ):
        lattice = pmg_Lattice( self.cell.matrix * self.scaling )
        structure = pmg_Structure( lattice, self.labels(), self.coordinates )
        return structure 

    @property
    def stoichiometry( self ):
        """
        Stoichiometry for this POSCAR, as a Counter.
        e.g. AB_2O_4 -> Counter( { 'A': 1, 'B': 2, O: 4 } )
        
        Args:
            None

        Returns:
            None
        """
        return Counter( { label: number for label, number in zip( self.atoms, self.atom_numbers ) } )



class Atom:
    """
    Class for individual atoms
    """

    def __init__( self, label, r ): # currently assume fractional coordinates
        """
        Initialise an Atom instance
    
        Args:
            label (Str): a label for this atom
            r (numpy.array): the atom coordinates
        
        Returns:
            None
        """
        self.label = label
        self.r = r



import numpy as np  # type: ignore
import re
from pymatgen.io.vasp.outputs import Outcar  # type: ignore

def reciprocal_lattice_from_outcar( filename ): # from https://github.com/MaterialsDiscovery/PyChemia
    """
    Finds and returns the reciprocal lattice vectors, if more than
    one set present, it just returns the last one.
    Args:
        filename (Str): The name of the outcar file to be read

    Returns:
        List(Float): The reciprocal lattice vectors.
    """
    outcar = open(filename, "r").read()
    # just keeping the last component
    recLat = re.findall(r"reciprocal\s*lattice\s*vectors\s*([-.\s\d]*)",
                        outcar)[-1]
    recLat = recLat.split()
    recLat = np.array(recLat, dtype=float)
    # up to now I have, both direct and rec. lattices (3+3=6 columns)
    recLat.shape = (3, 6)
    recLat = recLat[:, 3:]
    return recLat

def final_energy_from_outcar( filename='OUTCAR' ):
    """
    Finds and returns the energy from a VASP OUTCAR file, by searching for the last `energy(sigma->0)` entry.

    Args:
        filename (Str, optional): OUTCAR filename. Defaults to 'OUTCAR'.

    Returns:
        (Float): The last energy read from the OUTCAR file.
    """
    with open( filename ) as f:
        outcar = f.read()
    energy_re = re.compile( "energy\(sigma->0\) =\s+([-\d\.]+)" )
    energy = float( energy_re.findall( outcar )[-1] )
    return energy

def vasp_version_from_outcar( filename='OUTCAR' ):
    """
    Returns the first line from a VASP OUTCAR file, to get the VASP source version string.

    Args:
        filename (Str, optional): OUTCAR filename. Defaults to 'OUTCAR'.

    Returns:
        (Str): The first line read from the OUTCAR file.
    """
    with open( filename ) as f:
        line = f.readline().strip()
    return line

def potcar_eatom_list_from_outcar( filename='OUTCAR' ):
    """
    Returns a list of EATOM values for the pseudopotentials used.

    Args:
        filename (Str, optional): OUTCAR filename. Defaults to 'OUTCAR'.

    Returns:
        (List(Float)): A list of EATOM values, in the order they appear in the OUTCAR.
    """
    with open( filename ) as f:
        outcar = f.read()
    eatom_re = re.compile( "energy of atom\s+\d+\s+EATOM=\s*([-\d\.]+)" )
    eatom = [ float( e ) for e in eatom_re.findall( outcar ) ]
    return eatom


def fermi_energy_from_outcar( filename='OUTCAR' ):
    """Finds and returns the Fermi energy.

    Args:
        filename (:obj:'str', optional): the name of the ``OUTCAR`` file to be read. Default is `OUTCAR`.

    Returns:
        (Float): The Fermi energy as found in the ``OUTCAR`` file.

    """
    outcar = open(filename, "r").read()
    # returns a match object
    fermi_energy = re.search(r"E-fermi\s*:\s*([-.\d]*)", outcar)
    # take the first group - group(0) contains entire match
    fermi_energy = float(fermi_energy.group(1))
    return fermi_energy

def forces_from_outcar( filename='OUTCAR' ):
    """Finds and returns forces from the OUTCAR file.
      
    Args:
        filename (:obj:'str', optional): the name of the ``OUTCAR`` file to be read. Default is `OUTCAR`.

    Returns:
        (np.array): The force as found in the ``OUTCAR`` file, as a NSTEPS x NIONS x 3 numpy array.

    """
    outcar = Outcar(filename)
    forces = outcar.read_table_pattern(
        header_pattern=r"\sPOSITION\s+TOTAL-FORCE \(eV/Angst\)\n\s-+",
        row_pattern=r"\s+[+-]?\d+\.\d+\s+[+-]?\d+\.\d+\s+[+-]?\d+\.\d+\s+([+-]?\d+\.\d+)\s+([+-]?\d+\.\d+)\s+([+-]?\d+\.\d+)",
        footer_pattern=r"\s--+",
        postprocess=lambda x: float(x),
        last_one_only=False
    )
    return np.array( forces )

def coords_from_outcar( filename='OUTCAR' ):
    """Finds and returns Cartesian coordinates from the OUTCAR file.
      
    Args:
        filename (:obj:'str', optional): the name of the ``OUTCAR`` file to be read. Default is `OUTCAR`.

    Returns:
        (np.array): The Cartesian coordinates as found in the ``OUTCAR`` file, as a NSTEPS x NIONS x 3 numpy array.

    """
    outcar = Outcar(filename)
    coords = outcar.read_table_pattern(
        header_pattern=r"\sPOSITION\s+TOTAL-FORCE \(eV/Angst\)\n\s-+",
        row_pattern=r"\s+[+-]?(\d+\.\d+)\s+[+-]?(\d+\.\d+)\s+[+-]?(\d+\.\d+)\s+[+-]?\d+\.\d+\s+[+-]?\d+\.\d+\s+[+-]?\d+\.\d+",
        footer_pattern=r"\s--+",
        postprocess=lambda x: float(x),
        last_one_only=False
    )
    return np.array( coords )
      



import numpy as np  # type: ignore
import sys
import math
from vasppy import poscar, cell

def interpolate( i, j, x ):
    return( ( i * ( 1.0 - x ) ) + ( j * x) )

def trilinear_interpolation( cube, r ):
    return( interpolate ( 
                interpolate( 
                    interpolate( cube[ 0, 0, 0 ], cube[ 1, 0, 0 ], r[ 0 ] ), # trilinear interpolation => http://en.wikipedia.org/wiki/Trilinear_interpolation
                    interpolate( cube[ 0, 1, 0 ], cube[ 1, 1, 0 ], r[ 0 ] ), 
                    r[ 1 ] ),
                interpolate( 
                    interpolate( cube[ 0, 0, 1 ], cube[ 1, 0, 1 ], r[ 0 ] ),
                    interpolate( cube[ 0, 1, 1 ], cube[ 1, 1, 1 ], r[ 0 ] ), 
                    r[ 1 ] ), 
                r[ 2 ] ) 
            )

class Grid:

    projections = { 'x' : 0, 'y' : 1, 'z' : 2 }

    def __init__( self, dimensions = [ 1, 1, 1 ] ):
        self.filename = None
        self.poscar = poscar.Poscar()
        self.number_of_header_lines = 0
        self.dimensions = dimensions
        self.spacing = np.array( [ 1.0 / number_of_points for number_of_points in self.dimensions ] )
        self.grid = np.zeros( self.dimensions )

    def read_from_filename( self, filename ):
        self.filename = filename
        self.poscar = poscar.Poscar()
        self.poscar.read_from( self.filename )
        self.number_of_header_lines = sum( self.poscar.atom_numbers ) + poscar.Poscar.lines_offset
        self.read_dimensions()
        self.read_grid()
        return self

    def write_to_filename( self, filename ):
        with open( filename, 'w' ) as file_out:
            sys.stdout = file_out
            self.poscar.output()
            self.write_dimensions()
            sys.stdout.flush()
            self.write_grid()

    def read_dimensions( self ):
        with open( self.filename, 'r' ) as file_in:
            for i, line in enumerate( file_in ):
                if i == self.number_of_header_lines:
                    self.dimensions = [ int(i) for i in line.split() ]
                    break

    def write_dimensions( self ):
        print( "\n" + ' '.join( [ str(i) for i in self.dimensions ] ) ) 

    def read_grid( self ):
        grid_data = []
        grid_data_lines = math.ceil( ( self.dimensions[0] * self.dimensions[1] * self.dimensions[2] ) / 5 )
        with open( self.filename ) as file_in:
            for i, line in enumerate( file_in ):
                if ( i > self.number_of_header_lines ) and ( i <= self.number_of_header_lines + grid_data_lines ):
                    grid_data.append( line.strip() )
        grid_data = np.array( [ float( s ) for s in ' '.join( grid_data ).split() ] )
        self.grid = np.reshape( grid_data, tuple( self.dimensions ), order = 'F' )

    def write_grid( self ):
        np.savetxt( sys.stdout.buffer, np.swapaxes( self.grid, 0, 2 ).reshape( -1, 5 ), fmt='%.11E' )

    def average( self, normal_axis_label ):
        axes = [ 0, 1, 2 ]
        axes.remove( Grid.projections[ normal_axis_label ] )
        return( np.sum( np.sum( self.grid, axis=axes[1] ), axis=axes[0] ) / ( self.dimensions[0] * self.dimensions[1] ) )

    def by_index( self, index ):
        return self.grid[ index[0], index[1], index[2] ]

    def fractional_coordinate_at_index( self, index ):      
        return( np.multiply( self.spacing, index ) )

    def cartesian_coordinate_at_index( self, index ):
        return( self.fractional_coordinate_at_index( index ).dot( self.poscar.cell.matrix ) )

    def cube_slice( self, x0, y0, z0 ):
        x1 = ( x0 + 1 ) % self.dimensions[0]
        y1 = ( y0 + 1 ) % self.dimensions[1]
        z1 = ( z0 + 1 ) % self.dimensions[2]
        cube = np.array( [ self.grid[ x0, y0, z0 ],
                           self.grid[ x0, y0, z1 ],
                           self.grid[ x0, y1, z0 ],
                           self.grid[ x0, y1, z1 ],
                           self.grid[ x1, y0, z0 ],
                           self.grid[ x1, y0, z1 ],
                           self.grid[ x1, y1, z0 ],
                           self.grid[ x1, y1, z1 ] ] ).reshape( 2, 2, 2 )
        return( cube )

    def interpolated_value_at_fractional_coordinate( self, coord ):
        point = np.multiply( np.array( self.dimensions ), coord ) # point contains the (fractional) index of the coordinate coord.
        origin = [ int( f ) for f in point ]                      # origin contains the 3D index of the lowest-index point in the cube surrounding point (i,j,k)
        delta = [ p - o for p, o in zip( point, origin ) ]        # delta contains the *fractional* offset of "point" from "origin"
        cube = self.cube_slice( *origin )                         # cube contains the data values at the 8 bounding grid points
        return( trilinear_interpolation( cube, delta ) )

    def interpolate_to_orthorhombic_grid( self, dimensions ): # warning. This may need a more robust minimim image function in Cell.py for highly non-orthorhombic cells
        old_grid = self
        old_cell = old_grid.poscar.cell
        new_grid = Grid( dimensions = dimensions )
        new_grid.poscar.cell = cell.Cell( np.diag( np.diag( self.poscar.cell.matrix ) ) )
        index_grid           = np.array( [ [ i, j, k ] for ( i, j, k ), value in np.ndenumerate( new_grid.grid ) ] )
        cart_coord_grid      = np.array( [ new_grid.cartesian_coordinate_at_index( index ) for index in index_grid ] )
        init_frac_coord_grid = np.array( [ old_cell.cartesian_to_fractional_coordinates( r ) for r in cart_coord_grid ] )
        frac_coord_grid      = np.array( [ old_cell.inside_cell( r ) for r in init_frac_coord_grid ] )
        new_grid_data        = np.array( [ old_grid.interpolated_value_at_fractional_coordinate( r ) for r in frac_coord_grid ] )
        new_grid.grid = new_grid_data.reshape( new_grid.dimensions )
        return( new_grid )



import numpy as np  # type: ignore
from typing import Optional


class AutoKPoints:
    """
    class for automatic k-point generation data in KPOINTS.
    """

    def __init__(self,
                 title: str,
                 subdivisions: np.ndarray,
                 grid_centering: Optional[str] = 'G',
                 shift: Optional[np.ndarray] = np.array([0., 0., 0.])) -> None:
        """
        Initialise an AutoKPoints object

        Args:
            title (Str): The first line of the file, treated as a comment by VASP.
            subdivisions: (np.Array( Int, Int, Int )): Numbers of subdivisions along each reciprocal lattice vector.
            grid_centering (Str, optional): Specify gamma-centered (G) or the original Monkhorst-Pack scheme (MP). Default is 'G'.
            shift: (np.Array( Float, Float, Float ), optional): Optional shift of the mesh (s_1, s_2, s_3). Default is ( [ 0., 0., 0. ] ).

        Returns:
            None
        """
        accepted_grid_centerings = ['G', 'MP']
        if grid_centering not in accepted_grid_centerings:
            raise ValueError
        self.title = title
        self.grid_centering = grid_centering
        self.subdivisions = subdivisions
        self.shift = shift



"""
vasppy: Python utilities for working with VASP inputs and outputs.
"""

from setuptools import setup, find_packages
from vasppy import __version__ as VERSION

readme = 'README.md'
long_description = open(readme).read()

scripts = ['check_species',
           'murnfit', 
           'vasp_summary', 
           'poscar_to_cif', 
           'potcar_spec',
           'effective_mass',
           'fat_bands',
           'pimaim_to_poscar',
           'pimaim_to_xtl',
           'poscar_sort',
           'poscar_to_pimaim',
           'poscar_to_xtl',
           'proc_poscar',
           'rotate_poscar',
           'spacegroup',
           'vasp_grid',
           'xdatcar_to_disp',
           'xdatcar_to_poscart',
           'xdatcar_to_rdf']

setup(
    name='vasppy',
    version=VERSION,
    description='Python utilities for working with VASP inputs and outputs',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Benjamin J. Morgan',
    author_email='bjm42@bath.ac.uk',
    url='https://github.com/bjmorgan/vasppy', 
    download_url='https://github.com/bjmorgan/vasppy/archive/{}.tar.gz'.format(VERSION),
    keywords=['vasp'], # keywords
    packages=find_packages(exclude=['docs', 'tests*']),
    package_data={'vasppy': ['data/*.yaml']},
    entry_points={'console_scripts':[
                      '{} = vasppy.scripts.{}:main'.format(s, s) for s in scripts]},
    license='MIT',
    install_requires=['monty',
                      'numpy>=1.16.2',
                      'pandas',
                      'pymatgen',
                      'PyYAML', 
                      'coverage==4.3.4',
                      'codeclimate-test-reporter',
                      'fortranformat',
                      'scipy>=1.4.1', 
                      'tqdm',
                      'lxml'],
    python_requires='>=3.6'
    )



#! /usr/bin/env python3

from vasppy.summary import potcar_spec
import argparse

def parse_command_line_arguments():
    parser = argparse.ArgumentParser( description='Generate POTCAR specification based on hashing individual pseudopotential strings' )
    parser.add_argument('potcar', help="filename of the VASP POTCAR to be processed", nargs='?', default='POTCAR' )
    parser.add_argument('--hash', help="return the md5 hashes of the individual pseudopotential strings", action='store_true') 
    args = parser.parse_args()
    return args

def main():
    args = parse_command_line_arguments()
    if args.hash:
        hashes = {}
        for p, md5hash in potcar_spec(args.potcar, return_hashes=True).items():
            hashes[p] = md5hash
    for p, ps in potcar_spec(args.potcar).items():
        if args.hash:
            print(p, ps, hashes[p])
        else:
            print(p, ps, hashes[p])

if __name__ == '__main__':
    main()



import warnings

def handle_occupancy( occupancy, negative_occupancies='warn' ):
    valid_negative_occupancies = [ 'warn', 'raise', 'ignore', 'zero' ]
    if negative_occupancies not in valid_negative_occupancies:
        raise ValueError( "valid options for negative_occupancies are {}".format( valid_negative_occupancies ) )
    if occupancy < 0:
       if negative_occupancies == 'warn':
           warnings.warn( "One or more occupancies in your PROCAR file are negative." )
       elif negative_occupancies == 'raise':
           raise ValueError( "One or more occupancies in your PROCAR file are negative." )
       elif negative_occupancies == 'ignore':
           pass
       elif negative_occupancies == 'zero':
           occupancy = 0.0
    return occupancy

class Band():

    def __init__( self, index, energy, occupancy, negative_occupancies='warn' ):
        self.index = index
        self.energy = energy
        self.occupancy = handle_occupancy( occupancy, negative_occupancies=negative_occupancies )

    def __eq__( self, other ):
        return ( ( self.index == other.index ) &
                 ( self.energy == other.energy ) &
                 ( self.occupancy == other.occupancy ) )
